from typing import Iterable

from saengra.adapter import ShouldCommitAgain
from saengra.api import CommitStatus, Observation, Observer
from saengra.client import SaengraClient
from saengra.graph import Edge, Primitive, Subgraph, Update


class SocketAdapter:
    """
    SocketAdapter is the layer of abstraction between Python code and saengra-server commands.

    It communicates with saengra-server via a socket connection.
    It also maintains cache of incoming updates and flushes them when necessary.
    """

    def __init__(self, client: SaengraClient) -> None:
        self._client = client
        self._pending_updates: list[Update] = []

    def update(
        self, update_or_updates: Update | list[Update] | tuple[Update, ...]
    ) -> None:
        if isinstance(update_or_updates, (list, tuple)):
            self._pending_updates.extend(update_or_updates)
        else:
            self._pending_updates.append(update_or_updates)

    def flush(self) -> None:
        if not self._pending_updates:
            return
        self._client.apply_updates(self._pending_updates)
        self._pending_updates.clear()

    def commit(self) -> tuple[ShouldCommitAgain, list[Observation]]:
        self.flush()
        result = self._client.commit()
        return (
            result.status == CommitStatus.REQUIRES_REPEATED_COMMIT,
            result.observations,
        )

    def rollback(self) -> None:
        self._client.rollback()
        self._pending_updates.clear()

    def find_vertices(self, *, with_type_name: str | None = None) -> list[Primitive]:
        # Frequently needed in application-level code.
        self.flush()
        result = self._client.find(vertices=True, with_type_name=with_type_name)
        return result.vertices

    def find_edges(
        self,
        *,
        with_label: str | None = None,
        from_vertices: list[Primitive] | tuple[Primitive, ...] | None = None,
        to_vertices: list[Primitive] | tuple[Primitive, ...] | None = None,
    ) -> list[Edge]:
        self.flush()
        result = self._client.find(
            edges=True,
            with_label=with_label,
            from_vertices=from_vertices,
            to_vertices=to_vertices,
        )
        return result.edges

    def find_all(self) -> tuple[list[Primitive], list[Edge]]:
        # Needed to dump entire graph state.
        self.flush()
        result = self._client.find(vertices=True, edges=True)
        return result.vertices, result.edges

    def observe(self, observers: Iterable[Observer]) -> None:
        self.flush()
        self._client.observe(observers)

    def match(self, expression: str, *placeholder_values: Primitive) -> list[Subgraph]:
        self.flush()
        result = self._client.match(expression, placeholder_values)
        return result.subgraphs



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: messages.proto
# Protobuf Python Version: 6.33.0
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC, 6, 33, 0, "", "messages.proto"
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0emessages.proto\x12\x0bsaengra_api"*\n\x06Vertex\x12\x11\n\ttype_name\x18\x01 \x01(\x0c\x12\r\n\x05value\x18\x02 \x01(\x0c"Y\n\x04\x45\x64ge\x12!\n\x04\x66rom\x18\x01 \x01(\x0b\x32\x13.saengra_api.Vertex\x12\r\n\x05label\x18\x02 \x01(\x0c\x12\x1f\n\x02to\x18\x03 \x01(\x0b\x32\x13.saengra_api.Vertex"\x86\x01\n\x08Position\x12#\n\x06vertex\x18\x01 \x01(\x0b\x32\x13.saengra_api.Vertex\x12\x30\n\x04kind\x18\x02 \x01(\x0e\x32".saengra_api.Position.PositionKind"#\n\x0cPositionKind\x12\x08\n\x04\x43ORE\x10\x00\x12\t\n\x05ORBIT\x10\x01"\xa1\x02\n\x08Subgraph\x12-\n\x0estart_position\x18\x01 \x01(\x0b\x32\x15.saengra_api.Position\x12,\n\rend_positions\x18\x02 \x03(\x0b\x32\x15.saengra_api.Position\x12%\n\x08vertices\x18\x03 \x03(\x0b\x32\x13.saengra_api.Vertex\x12 \n\x05\x65\x64ges\x18\x04 \x03(\x0b\x32\x11.saengra_api.Edge\x12-\n\x04refs\x18\x05 \x03(\x0b\x32\x1f.saengra_api.Subgraph.RefsEntry\x1a@\n\tRefsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12"\n\x05value\x18\x02 \x01(\x0b\x32\x13.saengra_api.Vertex:\x02\x38\x01"\xb4\x02\n\x07Request\x12\'\n\x07\x63onnect\x18\x01 \x01(\x0b\x32\x14.saengra_api.ConnectH\x00\x12\x32\n\rapply_updates\x18\x02 \x01(\x0b\x32\x19.saengra_api.ApplyUpdatesH\x00\x12%\n\x06\x63ommit\x18\x03 \x01(\x0b\x32\x13.saengra_api.CommitH\x00\x12)\n\x08rollback\x18\x04 \x01(\x0b\x32\x15.saengra_api.RollbackH\x00\x12\'\n\x07observe\x18\x05 \x01(\x0b\x32\x14.saengra_api.ObserveH\x00\x12!\n\x04\x66ind\x18\x06 \x01(\x0b\x32\x11.saengra_api.FindH\x00\x12#\n\x05match\x18\x07 \x01(\x0b\x32\x12.saengra_api.MatchH\x00\x42\t\n\x07request"\x1b\n\x07\x43onnect\x12\x10\n\x08graph_id\x18\x01 \x01(\t"\xde\x02\n\x0c\x41pplyUpdates\x12\x31\n\x07updates\x18\x01 \x03(\x0b\x32 .saengra_api.ApplyUpdates.Update\x1a\x9a\x02\n\x06Update\x12\x39\n\x04kind\x18\x01 \x01(\x0e\x32+.saengra_api.ApplyUpdates.Update.UpdateKind\x12!\n\x04\x66rom\x18\x02 \x01(\x0b\x32\x13.saengra_api.Vertex\x12\x12\n\x05label\x18\x03 \x01(\tH\x00\x88\x01\x01\x12$\n\x02to\x18\x04 \x01(\x0b\x32\x13.saengra_api.VertexH\x01\x88\x01\x01"g\n\nUpdateKind\x12\x0e\n\nADD_VERTEX\x10\x00\x12\x0c\n\x08\x41\x44\x44_EDGE\x10\x01\x12\x11\n\rREMOVE_VERTEX\x10\x02\x12\x0f\n\x0bREMOVE_EDGE\x10\x03\x12\x17\n\x13REMOVE_EDGES_TO_ALL\x10\x04\x42\x08\n\x06_labelB\x05\n\x03_to"\x08\n\x06\x43ommit"\n\n\x08Rollback"\xd2\x01\n\x07Observe\x12\x30\n\tobservers\x18\x01 \x03(\x0b\x32\x1d.saengra_api.Observe.Observer\x1a\x94\x01\n\x08Observer\x12\n\n\x02id\x18\x01 \x01(\t\x12\x12\n\nexpression\x18\x02 \x01(\t\x12/\n\x12placeholder_values\x18\x03 \x03(\x0b\x32\x13.saengra_api.Vertex\x12\x11\n\ton_create\x18\x04 \x01(\x08\x12\x11\n\ton_change\x18\x05 \x01(\x08\x12\x11\n\ton_delete\x18\x06 \x01(\x08"\xc3\x01\n\x04\x46ind\x12\x10\n\x08vertices\x18\x01 \x01(\x08\x12\x1b\n\x0ewith_type_name\x18\x02 \x01(\tH\x00\x88\x01\x01\x12\r\n\x05\x65\x64ges\x18\x03 \x01(\x08\x12\x17\n\nwith_label\x18\x04 \x01(\tH\x01\x88\x01\x01\x12!\n\x04\x66rom\x18\x06 \x03(\x0b\x32\x13.saengra_api.Vertex\x12\x1f\n\x02to\x18\x05 \x03(\x0b\x32\x13.saengra_api.VertexB\x11\n\x0f_with_type_nameB\r\n\x0b_with_label"L\n\x05Match\x12\x12\n\nexpression\x18\x01 \x01(\t\x12/\n\x12placeholder_values\x18\x02 \x03(\x0b\x32\x13.saengra_api.Vertex"\xee\x02\n\x08Response\x12/\n\x07\x63onnect\x18\x01 \x01(\x0b\x32\x1c.saengra_api.ConnectResponseH\x00\x12:\n\rapply_updates\x18\x02 \x01(\x0b\x32!.saengra_api.ApplyUpdatesResponseH\x00\x12-\n\x06\x63ommit\x18\x03 \x01(\x0b\x32\x1b.saengra_api.CommitResponseH\x00\x12\x31\n\x08rollback\x18\x04 \x01(\x0b\x32\x1d.saengra_api.RollbackResponseH\x00\x12/\n\x07observe\x18\x05 \x01(\x0b\x32\x1c.saengra_api.ObserveResponseH\x00\x12)\n\x04\x66ind\x18\x06 \x01(\x0b\x32\x19.saengra_api.FindResponseH\x00\x12+\n\x05match\x18\x07 \x01(\x0b\x32\x1a.saengra_api.MatchResponseH\x00\x42\n\n\x08response"|\n\x0f\x43onnectResponse\x12\x33\n\x06status\x18\x01 \x01(\x0e\x32#.saengra_api.ConnectResponse.Status\x12\x0f\n\x07\x63reated\x18\x02 \x01(\x08"#\n\x06Status\x12\x06\n\x02OK\x10\x00\x12\x11\n\rNOT_CONNECTED\x10\x01"u\n\x14\x41pplyUpdatesResponse\x12\x38\n\x06status\x18\x01 \x01(\x0e\x32(.saengra_api.ApplyUpdatesResponse.Status"#\n\x06Status\x12\x06\n\x02OK\x10\x00\x12\x11\n\rNOT_CONNECTED\x10\x01"\x83\x04\n\x0e\x43ommitResponse\x12\x32\n\x06status\x18\x01 \x01(\x0e\x32".saengra_api.CommitResponse.Status\x12<\n\x0bobservation\x18\x02 \x03(\x0b\x32\'.saengra_api.CommitResponse.Observation\x1a\xbb\x02\n\x0bObservation\x12\x13\n\x0bobserver_id\x18\x01 \x01(\t\x12\x45\n\x04type\x18\x02 \x01(\x0e\x32\x37.saengra_api.CommitResponse.Observation.ObservationType\x12I\n\tvariables\x18\x03 \x03(\x0b\x32\x36.saengra_api.CommitResponse.Observation.VariablesEntry\x1a\x45\n\x0eVariablesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12"\n\x05value\x18\x02 \x01(\x0b\x32\x13.saengra_api.Vertex:\x02\x38\x01">\n\x0fObservationType\x12\r\n\tON_CREATE\x10\x00\x12\r\n\tON_CHANGE\x10\x01\x12\r\n\tON_DELETE\x10\x02"A\n\x06Status\x12\x06\n\x02OK\x10\x00\x12\x11\n\rNOT_CONNECTED\x10\x01\x12\x1c\n\x18REQUIRES_REPEATED_COMMIT\x10\x02"m\n\x10RollbackResponse\x12\x34\n\x06status\x18\x01 \x01(\x0e\x32$.saengra_api.RollbackResponse.Status"#\n\x06Status\x12\x06\n\x02OK\x10\x00\x12\x11\n\rNOT_CONNECTED\x10\x01"k\n\x0fObserveResponse\x12\x33\n\x06status\x18\x01 \x01(\x0e\x32#.saengra_api.ObserveResponse.Status"#\n\x06Status\x12\x06\n\x02OK\x10\x00\x12\x11\n\rNOT_CONNECTED\x10\x01"\xae\x01\n\x0c\x46indResponse\x12\x30\n\x06status\x18\x01 \x01(\x0e\x32 .saengra_api.FindResponse.Status\x12%\n\x08vertices\x18\x02 \x03(\x0b\x32\x13.saengra_api.Vertex\x12 \n\x05\x65\x64ges\x18\x03 \x03(\x0b\x32\x11.saengra_api.Edge"#\n\x06Status\x12\x06\n\x02OK\x10\x00\x12\x11\n\rNOT_CONNECTED\x10\x01"\x91\x01\n\rMatchResponse\x12\x31\n\x06status\x18\x01 \x01(\x0e\x32!.saengra_api.MatchResponse.Status\x12(\n\tsubgraphs\x18\x02 \x03(\x0b\x32\x15.saengra_api.Subgraph"#\n\x06Status\x12\x06\n\x02OK\x10\x00\x12\x11\n\rNOT_CONNECTED\x10\x01\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "messages_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_SUBGRAPH_REFSENTRY"]._loaded_options = None
    _globals["_SUBGRAPH_REFSENTRY"]._serialized_options = b"8\001"
    _globals["_COMMITRESPONSE_OBSERVATION_VARIABLESENTRY"]._loaded_options = None
    _globals["_COMMITRESPONSE_OBSERVATION_VARIABLESENTRY"]._serialized_options = (
        b"8\001"
    )
    _globals["_VERTEX"]._serialized_start = 31
    _globals["_VERTEX"]._serialized_end = 73
    _globals["_EDGE"]._serialized_start = 75
    _globals["_EDGE"]._serialized_end = 164
    _globals["_POSITION"]._serialized_start = 167
    _globals["_POSITION"]._serialized_end = 301
    _globals["_POSITION_POSITIONKIND"]._serialized_start = 266
    _globals["_POSITION_POSITIONKIND"]._serialized_end = 301
    _globals["_SUBGRAPH"]._serialized_start = 304
    _globals["_SUBGRAPH"]._serialized_end = 593
    _globals["_SUBGRAPH_REFSENTRY"]._serialized_start = 529
    _globals["_SUBGRAPH_REFSENTRY"]._serialized_end = 593
    _globals["_REQUEST"]._serialized_start = 596
    _globals["_REQUEST"]._serialized_end = 904
    _globals["_CONNECT"]._serialized_start = 906
    _globals["_CONNECT"]._serialized_end = 933
    _globals["_APPLYUPDATES"]._serialized_start = 936
    _globals["_APPLYUPDATES"]._serialized_end = 1286
    _globals["_APPLYUPDATES_UPDATE"]._serialized_start = 1004
    _globals["_APPLYUPDATES_UPDATE"]._serialized_end = 1286
    _globals["_APPLYUPDATES_UPDATE_UPDATEKIND"]._serialized_start = 1166
    _globals["_APPLYUPDATES_UPDATE_UPDATEKIND"]._serialized_end = 1269
    _globals["_COMMIT"]._serialized_start = 1288
    _globals["_COMMIT"]._serialized_end = 1296
    _globals["_ROLLBACK"]._serialized_start = 1298
    _globals["_ROLLBACK"]._serialized_end = 1308
    _globals["_OBSERVE"]._serialized_start = 1311
    _globals["_OBSERVE"]._serialized_end = 1521
    _globals["_OBSERVE_OBSERVER"]._serialized_start = 1373
    _globals["_OBSERVE_OBSERVER"]._serialized_end = 1521
    _globals["_FIND"]._serialized_start = 1524
    _globals["_FIND"]._serialized_end = 1719
    _globals["_MATCH"]._serialized_start = 1721
    _globals["_MATCH"]._serialized_end = 1797
    _globals["_RESPONSE"]._serialized_start = 1800
    _globals["_RESPONSE"]._serialized_end = 2166
    _globals["_CONNECTRESPONSE"]._serialized_start = 2168
    _globals["_CONNECTRESPONSE"]._serialized_end = 2292
    _globals["_CONNECTRESPONSE_STATUS"]._serialized_start = 2257
    _globals["_CONNECTRESPONSE_STATUS"]._serialized_end = 2292
    _globals["_APPLYUPDATESRESPONSE"]._serialized_start = 2294
    _globals["_APPLYUPDATESRESPONSE"]._serialized_end = 2411
    _globals["_APPLYUPDATESRESPONSE_STATUS"]._serialized_start = 2257
    _globals["_APPLYUPDATESRESPONSE_STATUS"]._serialized_end = 2292
    _globals["_COMMITRESPONSE"]._serialized_start = 2414
    _globals["_COMMITRESPONSE"]._serialized_end = 2929
    _globals["_COMMITRESPONSE_OBSERVATION"]._serialized_start = 2547
    _globals["_COMMITRESPONSE_OBSERVATION"]._serialized_end = 2862
    _globals["_COMMITRESPONSE_OBSERVATION_VARIABLESENTRY"]._serialized_start = 2729
    _globals["_COMMITRESPONSE_OBSERVATION_VARIABLESENTRY"]._serialized_end = 2798
    _globals["_COMMITRESPONSE_OBSERVATION_OBSERVATIONTYPE"]._serialized_start = 2800
    _globals["_COMMITRESPONSE_OBSERVATION_OBSERVATIONTYPE"]._serialized_end = 2862
    _globals["_COMMITRESPONSE_STATUS"]._serialized_start = 2864
    _globals["_COMMITRESPONSE_STATUS"]._serialized_end = 2929
    _globals["_ROLLBACKRESPONSE"]._serialized_start = 2931
    _globals["_ROLLBACKRESPONSE"]._serialized_end = 3040
    _globals["_ROLLBACKRESPONSE_STATUS"]._serialized_start = 2257
    _globals["_ROLLBACKRESPONSE_STATUS"]._serialized_end = 2292
    _globals["_OBSERVERESPONSE"]._serialized_start = 3042
    _globals["_OBSERVERESPONSE"]._serialized_end = 3149
    _globals["_OBSERVERESPONSE_STATUS"]._serialized_start = 2257
    _globals["_OBSERVERESPONSE_STATUS"]._serialized_end = 2292
    _globals["_FINDRESPONSE"]._serialized_start = 3152
    _globals["_FINDRESPONSE"]._serialized_end = 3326
    _globals["_FINDRESPONSE_STATUS"]._serialized_start = 2257
    _globals["_FINDRESPONSE_STATUS"]._serialized_end = 2292
    _globals["_MATCHRESPONSE"]._serialized_start = 3329
    _globals["_MATCHRESPONSE"]._serialized_end = 3474
    _globals["_MATCHRESPONSE_STATUS"]._serialized_start = 2257
    _globals["_MATCHRESPONSE_STATUS"]._serialized_end = 2292
# @@protoc_insertion_point(module_scope)



from itertools import chain
from typing import TYPE_CHECKING, Any, Iterable, Iterator

from saengra.utilities.typing import is_entity

if TYPE_CHECKING:
    from saengra.entity import EnvProtocol, Entity

from saengra.graph import (
    AddEdge,
    AddVertex,
    Label,
    Primitive,
    RemoveEdge,
    RemoveEdgesToAll,
    Update,
)


class RelatedEntitiesContainer:
    __slots__ = (
        "_env",
        "_e",
        "_from",
        "_label",
        "_cached_values",
    )

    def __init__(
        self,
        env: "EnvProtocol",
        e: "Entity",
        from_: Primitive,
        label: Label,
    ) -> None:
        self._env = env
        self._e = e
        self._from = from_
        self._label = label
        self._cached_values: set["Primitive | Entity"] = set()

    def assign(self, value: list["Primitive | Entity"]):
        self._clear()
        self._add_many(value)

    def _add_inverse_one(self, item: "Primitive | Entity"):
        if is_entity(item):
            item._inverse.add((self._e, self._label))

    def _add_inverse_many(self, items: list["Primitive | Entity"]):
        for item in items:
            if is_entity(item):
                item._inverse.add((self._e, self._label))

    def _remove_inverse(
        self, items: list["Primitive | Entity"] | set["Primitive | Entity"]
    ):
        for value in items:
            if is_entity(value):
                value._inverse.discard((self._e, self._label))

    def _add(self, item: "Primitive | Entity") -> None:
        if item is None:
            raise ValueError("expected primitive or entity, got None")

        if is_entity(item):
            to = item.primitive
            self._env.add(item)
        else:
            to = item
            self._env.update(AddVertex(to))

        self._env.update(AddEdge(self._from, self._label, to))
        self._add_to_cache(item)

    def _add_to_cache(self, item: "Primitive | Entity") -> None:
        self._cached_values.add(item)
        self._add_inverse_one(item)

    def _add_many(
        self, items: list["Primitive | Entity"] | set["Primitive | Entity"]
    ) -> None:
        updates: list[AddVertex | AddEdge] = []
        for item in items:
            if is_entity(item):
                if not item.alive:
                    raise RuntimeError(
                        f"trying to assign `.{self._label}` to [..., {item}, ...] that is not alive"
                    )
                updates.append(AddEdge(self._from, self._label, item.primitive))
            else:
                updates.append(AddVertex(item))
                updates.append(AddEdge(self._from, self._label, item))
        self._env.update_many(updates)
        self._cached_values.update(items)
        self._add_inverse_many(items)

    def _discard(self, item: "Primitive | Entity") -> None:
        to = item.primitive if is_entity(item) else item
        self._env.update(RemoveEdge(self._from, self._label, to))
        self._cached_values.discard(item)
        self._remove_inverse([item])

    def _discard_many(
        self, items: list["Primitive | Entity"] | set["Primitive | Entity"]
    ) -> None:
        remove_edge_updates: list[Update] = [
            RemoveEdge(
                self._from, self._label, item.primitive if is_entity(item) else item
            )
            for item in items
        ]
        self._env.update_many(remove_edge_updates)
        self._cached_values.difference_update(items)
        self._remove_inverse(items)

    def _clear(self) -> None:
        self._env.update(RemoveEdgesToAll(self._from, self._label))
        self._remove_inverse(self._cached_values)
        self._cached_values.clear()

    def __len__(self) -> int:
        return len(self._cached_values)

    def __bool__(self) -> bool:
        return bool(self._cached_values)

    def __contains__(self, o: object) -> bool:
        return o in self._cached_values

    def __iter__(self) -> Iterator[Primitive]:
        return iter(self._cached_values)


class RelatedEntitiesSet(RelatedEntitiesContainer):
    def __repr__(self) -> str:
        return repr(self._cached_values)

    def add(self, item: "Primitive | Entity") -> None:
        self._add(item)

    def discard(self, item: "Primitive | Entity") -> None:
        self._discard(item)

    def clear(self) -> None:
        self._clear()

    def update(self, *s: Iterable["Primitive | Entity"]) -> None:
        if len(s) == 1:
            # Shorter path for a frequent use.
            items = s[0]
            if items is self:
                # This is a necessary check. We might hurt ourselves if
                # performed _add_many() with self as an argument,
                # which is going to change during the function execution.
                return
            if type(items) not in (list, set, RelatedEntitiesSet):
                items = list(items)
            self._add_many(items)
            return

        items = list(chain.from_iterable(s))
        self._add_many(items)

    def difference_update(self, *s: Iterable["Primitive | Entity"]) -> None:
        if len(s) == 1:
            # Shorter path for a frequent use.
            items = s[0]
            if items is self:
                self._clear()
                return
            if type(items) not in (list, set, RelatedEntitiesSet):
                items = list(items)
            self._discard_many(items)
            return

        items = list(chain.from_iterable(s))
        self._discard_many(items)

    def __eq__(self, other: Any) -> bool:
        match other:
            case set():
                return self._cached_values == other
            case RelatedEntitiesSet():
                return self._cached_values == other._cached_values
        return NotImplemented

    def __and__(self, other):
        return self._cached_values & other

    def __rand__(self, other):
        return other & self._cached_values

    def __or__(self, other):
        return self._cached_values | other

    def __ror__(self, other):
        return other | self._cached_values

    def __sub__(self, other):
        return self._cached_values - other

    def __rsub__(self, other):
        return other - self._cached_values

    def __ge__(self, other):
        return self._cached_values >= other

    def __gt__(self, other):
        return self._cached_values > other

    def __le__(self, other):
        return self._cached_values <= other

    def __lt__(self, other):
        return self._cached_values < other



from dataclasses import dataclass, field
from typing import Hashable, Literal, TypeAlias

from saengra.frozendict import frozendict

Primitive: TypeAlias = Hashable
Vertex: TypeAlias = Primitive
Label: TypeAlias = str


# Store reference to the function in module globals for maximum performance.
_object_setattr = object.__setattr__


@dataclass(frozen=True, slots=True)
class Edge:
    from_: Vertex
    label: Label
    to: Vertex

    @property
    def reverse(self) -> "Edge":
        return Edge(from_=self.to, label=self.label, to=self.from_)

    # Manual implementations are way faster compared to built-in _dataclass_getstate()/_dataclass_setstate().
    # And since edges can be frequently pickled (e. g. to save graph snapshot), it's better to have this here.
    def __getstate__(self) -> tuple[Vertex, Label, Vertex]:
        return self.from_, self.label, self.to

    def __setstate__(self, state: tuple[Vertex, Label, Vertex]) -> None:
        _object_setattr(self, "from_", state[0])
        _object_setattr(self, "label", state[1])
        _object_setattr(self, "to", state[2])


def reverse(edge: Edge) -> Edge:
    return Edge(from_=edge.to, label=edge.label, to=edge.from_)


@dataclass(frozen=True, slots=True)
class AddVertex:
    primitive: Primitive


@dataclass(frozen=True, slots=True)
class AddEdge:
    from_: Primitive
    label: Label
    to: Primitive


@dataclass(frozen=True, slots=True)
class RemoveVertex:
    primitive: Primitive


@dataclass(frozen=True, slots=True)
class RemoveEdge:
    from_: Primitive
    label: Label
    to: Primitive


@dataclass(frozen=True, slots=True)
class RemoveEdgesToAll:
    from_: Primitive
    label: Label


Update = AddVertex | AddEdge | RemoveVertex | RemoveEdge | RemoveEdgesToAll


@dataclass(frozen=True, slots=True)
class Position:
    vertex: Vertex
    point: Literal[".", "o"]


Refs: TypeAlias = frozendict[str, Vertex]


@dataclass(frozen=True, slots=True)
class Subgraph:
    start_position: Position
    end_positions: frozenset[Position]
    vertices: frozenset[Vertex] = field(default=frozenset())
    edges: frozenset[Edge] = field(default=frozenset())
    refs: Refs = field(default=Refs())



from functools import cache
from time import perf_counter
from typing import Iterable, Type, TypeVar
from weakref import WeakValueDictionary

from saengra.adapter import Adapter
from saengra.api import Observation
from saengra.c_extension import DirectAdapter
from saengra.client import SaengraClient
from saengra.entity import Entity, EnvProtocol
from saengra.graph import AddVertex, Edge, Primitive, RemoveVertex, Update
from saengra.observer import Observer, RefsHandler
from saengra.socket_adapter import SocketAdapter
from saengra.utilities.colors import light_green
from saengra.utilities.itertools import group_by
from saengra.utilities.loggers import LazyRefsFormatter, logger

T = TypeVar("T")


class Environment(EnvProtocol):
    def __init__(
        self,
        *,
        client: SaengraClient | None = None,
        adapter: Adapter | None = None,
        entity_types: Iterable[Type[Entity]] | None = None,
    ) -> None:
        self._entity_types = [] if entity_types is None else [*entity_types]
        self._primitive_to_entity = {
            entity_type.primitive_type: entity_type
            for entity_type in self._entity_types
        }
        self._currently_committing = False

        if adapter is not None:
            self._adapter = adapter
        elif client is not None:
            self._adapter = SocketAdapter(client=client)
        else:
            self._adapter = DirectAdapter()
        # self._profiler = Profiler()
        self._observers: dict[str, Observer] = {}

        self._entities: dict[Primitive, Entity] = {}
        self._temporary_entities: WeakValueDictionary[Primitive, Entity] = (
            WeakValueDictionary()
        )
        """ 
        Temporary entities are entities that refer to primitives that are not in the graph.
        E.g. in the following code

              e = MyEntity(graph=g, ...)
              if not e.alive:
                  env.add(e)
                  ...

        a temporary entity will be created in the line 1 which will be converted to a normal 
        entity in the line 3.
        """

    def add(self, *entities: Entity) -> None:
        updates: list[Update] = []
        for entity in entities:
            if entity.alive:
                continue
            updates.append(AddVertex(entity.primitive))
            entity._after_add()
            self._entities[entity.primitive] = entity
            del self._temporary_entities[entity.primitive]
        self._adapter.update(updates)

    def remove(self, *entities: Entity) -> None:
        updates: list[Update] = []
        for entity in entities:
            if not entity.alive:
                continue
            updates.append(RemoveVertex(entity.primitive))
            entity._after_remove()
            del self._entities[entity.primitive]
            self._temporary_entities[entity.primitive] = entity
        self._adapter.update(updates)

    def update(self, *updates: Update) -> None:
        """
        Apply low-level updates to the graph manually.

        Not compatible with Entity manipulation, so use at own risk.
        """
        self._adapter.update(updates)

    def update_many(self, updates: list[Update]) -> None:
        self._adapter.update(updates)

    def commit(self) -> None:
        if self._currently_committing:
            raise RuntimeError("should not invoke .commit() in pre-commit hooks")
        try:
            self._currently_committing = True
            self._do_commit()
        except Exception:
            self.rollback()
            raise
        finally:
            self._currently_committing = False

    def rollback(self) -> None:
        self._adapter.rollback()
        self.sync()

    def sync(self) -> None:
        """Synchronize entities with current graph state, probably after rollback."""
        self._entities.clear()
        self._temporary_entities.clear()

        vertices, edges = self._adapter.find_all()

        # Create entity objects
        for vertex in vertices:
            entity = self._wrap_into_new_entity(vertex)
            if isinstance(entity, Entity):
                entity._after_add()
                self._entities[entity.primitive] = entity
                del self._temporary_entities[entity.primitive]

        # Fill entity relations
        vertex_to_edges = group_by(edges, key=lambda e: e.from_)
        for entity in self._entities.values():
            edges = vertex_to_edges.get(entity.primitive, [])
            labels_and_tos = [
                (edge.label, self._wrap_into_existing_entity(edge.to)) for edge in edges
            ]
            entity._sync(edges=labels_and_tos)

    def find_all(self) -> tuple[list[Primitive], list[Edge]]:
        return self._adapter.find_all()

    def find_edges(
        self,
        *,
        with_label: str | None = None,
        to: Primitive | Entity | list[Primitive | Entity] | None = None,
    ) -> list[Edge]:
        container = [] if to is None else to if isinstance(to, (list, tuple)) else [to]
        edges = self._adapter.find_edges(
            with_label=with_label, to_vertices=[_unwrap_primitive(v) for v in container]
        )
        return edges

    def find_all_of_type(self, type_: Type[T]) -> set[T]:
        primitive_type = _get_primitive_type(type_)
        vertices = self._adapter.find_vertices(with_type_name=primitive_type.__name__)
        return {
            self._wrap_into_existing_entity(v)
            for v in vertices
            if isinstance(v, primitive_type)
        }

    def match(
        self, expression: str, *placeholder_values: Primitive
    ) -> list[dict[str, Primitive | Entity]]:
        subgraphs = self._adapter.match(expression, *placeholder_values)
        return [
            {k: self._wrap_into_existing_entity(v) for k, v in sg.refs.items()}
            for sg in subgraphs
        ]

    def register_observers(self, observers: list[Observer]) -> None:
        new_observers: dict[str, Observer] = {}
        for w in observers:
            if w.id in self._observers:
                raise RuntimeError(
                    f"duplicate observer ID: {w.id}\n\n"
                    "To avoid this error, register observers all at once, or make an extra ID deduplication step"
                )
            new_observers[w.id] = w
        self._observers.update(new_observers)
        self._adapter.observe(w.to_api() for w in observers)

    def _get_living_entity(self, primitive: Primitive) -> Entity | None:
        return self._entities.get(primitive, None)

    def _get_temporary_entity(self, primitive: Primitive) -> Entity | None:
        return self._temporary_entities.get(primitive, None)

    def _register_temporary_entity(
        self, primitive: Primitive, entity: "Entity"
    ) -> None:
        self._temporary_entities[primitive] = entity

    def _do_commit(self) -> None:
        before = perf_counter()
        should_commit_again = True
        num_iterations = 0
        while should_commit_again:
            should_commit_again, observations = self._adapter.commit()
            for o in observations:
                self._invoke_handlers(o)

            num_iterations += 1
            if should_commit_again and num_iterations > 100:
                raise RuntimeError("stuck in pre-commit loop")
        after = perf_counter()
        logger.info(
            f"Commit took {num_iterations} iterations, {after - before:.3f} seconds"
        )

    def _invoke_handlers(self, o: Observation) -> None:
        observer = self._observers[o.observer_id]
        handlers = observer.pick_handlers(o.type)
        for handler in handlers:
            self._invoke_handler(handler, o.variables)

    def _invoke_handler(self, handler: RefsHandler, refs: dict[str, Primitive]) -> None:
        logger.debug(
            f"Invoking {light_green(handler.__name__)}(%s)", LazyRefsFormatter(refs)
        )
        enriched_refs: dict[str, Primitive | Entity | Environment] = {}
        if env_arg := _get_env_arg(handler):
            enriched_refs[env_arg] = self
        for ref, primitive in refs.items():
            if entity_type := self._primitive_to_entity.get(type(primitive)):
                enriched_refs[ref] = entity_type(self, primitive)
            else:
                enriched_refs[ref] = primitive
        handler(**enriched_refs)

    def _wrap_into_new_entity(self, v: Primitive) -> Primitive | Entity:
        if type_ := self._primitive_to_entity.get(type(v)):
            return type_(env=self, primitive=v)
        return v

    def _wrap_into_existing_entity(self, v: Primitive) -> Primitive | Entity:
        if type_ := self._primitive_to_entity.get(type(v)):
            # Corresponding entity is supposed to exist.
            return self._entities[v]
        return v


def _unwrap_primitive(v: Primitive | Entity) -> Primitive:
    if isinstance(v, Entity):
        return v.primitive
    return v


def _get_primitive_type(t: Type) -> Type:
    if issubclass(t, Entity):
        return t.primitive_type
    return t


@cache
def _get_env_arg(handler: RefsHandler) -> str | None:
    for key, value in handler.__annotations__.items():
        if value is Environment:
            return key
    return None



"""High-level Saengra client."""

import os
import socket
import struct
import subprocess
import sys
import tempfile
import time
from pathlib import Path
from typing import Iterable

from saengra import messages_pb2
from saengra.api import CommitResponse, FindResponse, MatchResponse, Observer
from saengra.conversions import Converter
from saengra.errors import CommitError, ConnectError
from saengra.errors import ConnectionError as SaengraConnectionError
from saengra.errors import (
    FindError,
    MatchError,
    ObserveError,
    RollbackError,
    UpdateError,
)
from saengra.graph import Primitive, Update

# from saengra_pyproto import Converter as ConverterV2


def find_server_binary() -> str:
    if subprocess.call(["which", "saengra_server"]) == 0:
        return "saengra_server"

    project_root = Path(__file__).parent.parent
    candidates = [project_root / "saengra-server" / "build" / "saengra_server"]
    for candidate in candidates:
        if candidate.exists():
            return str(candidate)
    raise FileNotFoundError(f"could not find saengra-server binary")


class SaengraClient:
    """High-level client for Saengra graph database.

    Can either start saengra-server in the background or connect to an existing
    server via Unix domain socket.
    """

    def __init__(
        self,
        server_binary: str | None = None,
        socket_path: str | None = None,
    ):
        """Initialize the client.

        Args:
            server_binary: Path to saengra-server binary (default: "saengra-server" from PATH).
                          Set to None if connecting to existing socket.
            socket_path: Path to existing Unix socket. If provided, connects to this socket
                        instead of starting a new server process.
        """
        self.server_binary = server_binary or find_server_binary()
        self.socket_path: str | None = socket_path
        self.process: subprocess.Popen | None = None
        self.sock: socket.socket | None = None
        self._owns_server = socket_path is None  # Track if we started the server
        self._converter = Converter()
        # self._converter_v2 = ConverterV2()

    def __enter__(self):
        """Start the server and connect."""
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Stop the server and cleanup."""
        self.stop()

    def start(self):
        """Start the saengra-server process or connect to existing socket."""
        if self._owns_server:
            # Start a new server process
            if not self.server_binary:
                raise ValueError(
                    "server_binary must be provided when not using existing socket"
                )

            # Create temporary socket file
            fd, self.socket_path = tempfile.mkstemp(suffix=".sock")
            os.close(fd)
            os.unlink(self.socket_path)  # Server will create it

            # Start server process
            self.process = subprocess.Popen(
                [self.server_binary, self.socket_path],
                stdout=sys.stdout,
                stderr=sys.stderr,
                env={**os.environ, "SAENGRA_LOGS_COLOR": "1"},
            )

            max_wait = 5.0
            wait_interval = 0.1
            elapsed = 0.0
            while not os.path.exists(self.socket_path) and elapsed < max_wait:
                if self.process.poll() is not None:
                    raise RuntimeError("server process exited unexpectedly")
                time.sleep(wait_interval)
                elapsed += wait_interval

            if not os.path.exists(self.socket_path):
                raise RuntimeError(f"Server socket not created after {max_wait}s")
        else:
            # Connect to existing socket
            if not self.socket_path:
                raise ValueError(
                    "socket_path must be provided when connecting to existing server"
                )
            if not os.path.exists(self.socket_path):
                raise FileNotFoundError(
                    f"Socket file does not exist: {self.socket_path}"
                )

        # Connect to socket
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

        try:
            self.sock.connect(self.socket_path)
        except Exception as exc:
            raise SaengraConnectionError(
                f"failed to connect to {self.socket_path}"
            ) from exc

    def stop(self):
        """Stop the server process and cleanup."""
        if self.sock:
            try:
                self.sock.close()
            except Exception:
                pass
            self.sock = None

        # Only stop the server and clean up socket if we own it
        if self._owns_server:
            if self.process:
                self.process.terminate()
                try:
                    self.process.wait(timeout=5.0)
                except subprocess.TimeoutExpired:
                    self.process.kill()
                    self.process.wait()
                self.process = None

            if self.socket_path and os.path.exists(self.socket_path):
                try:
                    os.unlink(self.socket_path)
                except Exception:
                    pass
                self.socket_path = None

    def _send_request(
        self, request: messages_pb2.Request | bytes
    ) -> messages_pb2.Response:
        """Send a request and receive a response.

        Args:
            request: Request message to send

        Returns:
            Response message from server
        """
        if not self.sock:
            raise RuntimeError("Client not connected. Call start() first.")

        # Serialize request
        request_data = (
            request if isinstance(request, bytes) else request.SerializeToString()
        )

        # Send length-prefixed message (4 bytes for length, little-endian)
        length = len(request_data)
        self.sock.sendall(struct.pack("<I", length))
        self.sock.sendall(request_data)

        # Receive response length
        length_bytes = self._recv_exact(4)
        response_length = struct.unpack("<I", length_bytes)[0]

        # Receive response data
        response_data = self._recv_exact(response_length)

        # Parse response
        response = messages_pb2.Response()
        response.ParseFromString(response_data)
        return response

    def _recv_exact(self, num_bytes: int) -> bytes:
        """Receive exact number of bytes from socket.

        Args:
            num_bytes: Number of bytes to receive

        Returns:
            Received bytes
        """
        data = b""
        while len(data) < num_bytes:
            chunk = self.sock.recv(num_bytes - len(data))
            if not chunk:
                raise SaengraConnectionError("Socket connection closed")
            data += chunk
        return data

    def connect(self, graph_id: str) -> bool:
        """Connect to a graph (creates if doesn't exist).

        Args:
            graph_id: ID of the graph to connect to

        Returns:
            True if graph was created, False if it already existed

        Raises:
            ConnectError: If connection fails
        """
        request = messages_pb2.Request()
        request.connect.graph_id = graph_id
        response = self._send_request(request)

        if response.connect.status != messages_pb2.ConnectResponse.OK:
            raise ConnectError(f"Failed to connect to graph '{graph_id}'")

        return response.connect.created

    def apply_updates(self, updates: list[Update]) -> None:
        """Apply a sequence of updates to the graph.

        Args:
            updates: list of updates to apply

        Raises:
            UpdateError: If updates fail to apply
        """
        request = messages_pb2.Request()
        updates_proto = request.apply_updates.updates
        for u in updates:
            self._converter.update_to_proto(u, updates_proto.add())
        request_v1 = request.SerializeToString()

        # request_v2 = self._converter_v2.serialize_apply_updates_request(updates)

        response = self._send_request(request_v1)

        if response.apply_updates.status != messages_pb2.ApplyUpdatesResponse.OK:
            raise UpdateError("Failed to apply updates")

    def commit(self) -> CommitResponse:
        """Commit pending changes.

        Returns:
            CommitResponse containing status and observations

        Raises:
            CommitError: If commit fails
        """
        request = messages_pb2.Request()
        request.commit.CopyFrom(messages_pb2.Commit())
        response = self._send_request(request)

        if response.commit.status not in (
            messages_pb2.CommitResponse.OK,
            messages_pb2.CommitResponse.REQUIRES_REPEATED_COMMIT,
        ):
            raise CommitError("Failed to commit changes")

        return self._converter.commit_response_from_proto(response.commit)

    def rollback(self) -> None:
        """Rollback pending changes.

        Raises:
            RollbackError: If rollback fails
        """
        request = messages_pb2.Request()
        request.rollback.CopyFrom(messages_pb2.Rollback())
        response = self._send_request(request)

        if response.rollback.status != messages_pb2.RollbackResponse.OK:
            raise RollbackError("Failed to rollback changes")

    def observe(self, observers: Iterable[Observer]) -> None:
        """Register observers for expressions.

        Args:
            observers: list of observers to register

        Raises:
            ObserveError: If registration fails
        """
        request = messages_pb2.Request()
        proto_observers = request.observe.observers
        for o in observers:
            self._converter.observer_to_proto(o, proto_observers.add())
        response = self._send_request(request)

        if response.observe.status != messages_pb2.ObserveResponse.OK:
            raise ObserveError("Failed to register observers")

    def find(
        self,
        vertices: bool = False,
        edges: bool = False,
        with_type_name: str | None = None,
        with_label: str | None = None,
        from_vertices: list[Primitive] | tuple[Primitive, ...] | None = None,
        to_vertices: list[Primitive] | tuple[Primitive, ...] | None = None,
    ) -> FindResponse:
        """Find vertices and/or edges in the graph.

        Args:
            vertices: If True, return all vertices (or filtered by type_name if provided)
            edges: If True, return all edges
            with_type_name: Optional filter to only return vertices of this type
            with_label: Optional filter to only return edges with this label
            from_vertices: Optional filter to only return edges from these vertices
            to_vertices: Optional filter to only return edges to these vertices

        Returns:
            FindResponse with lists of vertices and edges

        Raises:
            FindError: If find operation fails
        """
        request = messages_pb2.Request()
        request.find.vertices = vertices
        request.find.edges = edges
        if with_type_name is not None:
            request.find.with_type_name = with_type_name
        if with_label is not None:
            request.find.with_label = with_label
        for vertex in from_vertices or []:
            proto_vertex = getattr(request.find, "from").add()
            self._converter.vertex_to_proto(vertex, proto_vertex)
        for vertex in to_vertices or []:
            proto_vertex = request.find.to.add()
            self._converter.vertex_to_proto(vertex, proto_vertex)

        response = self._send_request(request)

        if response.find.status != messages_pb2.FindResponse.OK:
            raise FindError("Failed to find vertices/edges")

        return self._converter.find_response_from_proto(response.find)

    def match(
        self,
        expression: str,
        placeholder_values: list[Primitive] | tuple[Primitive, ...] | None = None,
    ) -> MatchResponse:
        """Match graph patterns using an expression.

        Args:
            expression: Pattern expression to match
            placeholder_values: Optional list of vertex values for placeholders in the expression

        Returns:
            MatchResponse with list of matching subgraphs

        Raises:
            MatchError: If match operation fails
        """
        request = messages_pb2.Request()
        request.match.expression = expression
        if placeholder_values:
            proto_placeholder_values = request.match.placeholder_values
            for v in placeholder_values:
                self._converter.vertex_to_proto(v, proto_placeholder_values.add())

        response = self._send_request(request)

        if response.match.status != messages_pb2.MatchResponse.OK:
            raise MatchError("Failed to match pattern")

        return self._converter.match_response_from_proto(response.match)



from enum import Enum
from typing import Any, TypeVar, get_args

from frozendict import frozendict as fd

try:
    from pydantic import GetCoreSchemaHandler
    from pydantic_core import CoreSchema
    from pydantic_core.core_schema import (
        dict_schema,
        is_instance_schema,
        no_info_after_validator_function,
        str_schema,
    )
except ImportError:
    # No pydantic!
    GetCoreSchemaHandler = None  # type: ignore[assignment,misc]

K = TypeVar("K")
V = TypeVar("V")


class frozendict(fd[K, V]):

    if GetCoreSchemaHandler:  # type: ignore[truthy-function]

        @classmethod
        def __get_pydantic_core_schema__(
            cls, source: Any, handler: GetCoreSchemaHandler
        ) -> CoreSchema:
            args = get_args(source)
            if not args or len(args) != 2:
                return is_instance_schema(cls)
            key_type, value_type = args
            value_schema = handler.generate_schema(value_type)
            schema = dict_schema(keys_schema=str_schema(), values_schema=value_schema)

            def convert_keys(kv: dict) -> "frozendict":
                if isinstance(key_type, type) and issubclass(key_type, Enum):
                    return cls({key_type(k): v for k, v in kv.items()})
                return cls(kv)

            return no_info_after_validator_function(convert_keys, schema)

    def set(self, key, val):
        new_self = dict(self)
        new_self[key] = val
        return self.__class__(new_self)

    def setdefault(self, key, default=None):
        if key in self:
            return self
        new_self = dict(self)
        new_self[key] = default
        return self.__class__(new_self)

    def discard(self, *keys):
        new_self = dict(self)
        for key in keys:
            new_self.pop(key, None)
        if new_self:
            return self.__class__(new_self)
        return self.__class__()

    def delete(self, *keys):
        new_self = dict(self)
        for key in keys:
            del new_self[key]
        if new_self:
            return self.__class__(new_self)
        return self.__class__()



import logging
import sys

from saengra.graph import Primitive
from saengra.utilities.colors import bold_red, cyan, green, red, yellow

COLORS = {
    "DEBUG": cyan,
    "INFO": green,
    "WARNING": yellow,
    "ERROR": red,
    "CRITICAL": bold_red,
}


class ColoredFormatter(logging.Formatter):
    def format(self, record):
        levelname = record.levelname
        if levelname in COLORS:
            record.levelname = f"{COLORS[levelname](levelname)}"
        return super().format(record)


def configure_logging():
    """Configure the 'saengra' logger to output all records (DEBUG+) to stdout."""
    logger = logging.getLogger("saengra")
    logger.setLevel(logging.DEBUG)

    # Remove existing handlers to avoid duplicates
    logger.handlers.clear()

    # Create stdout handler
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)

    # Set format
    formatter = ColoredFormatter(
        "[saengra] [%(asctime)s.%(msecs)03d] [%(levelname)s] %(message)s",
        datefmt="%H:%M:%S",
    )
    handler.setFormatter(formatter)

    logger.addHandler(handler)

    # Prevent propagation to root logger
    logger.propagate = False


logger = logging.getLogger("saengra")


class LazyRefsFormatter:
    """Optimize logging calls in production environment by not forming
    strings unless the logging level demands it."""

    __slots__ = ("_refs",)

    def __init__(self, refs: dict[str, Primitive]) -> None:
        self._refs = refs

    def __str__(self) -> str:
        return ", ".join(f"{k}={v!r}" for k, v in sorted(self._refs.items()))



import pytest

from saengra import Environment
from saengra.graph import (
    AddVertex,
    AddEdge,
    Edge,
    RemoveVertex,
    RemoveEdge,
    RemoveEdgesToAll,
)
from tests.scenarios.game import Biome


def test_add_vertex_unhashable(empty_env: Environment) -> None:
    with pytest.raises(TypeError):
        empty_env.update(AddVertex([]))


def test_add_vertex_unpicklable(empty_env: Environment) -> None:
    class Unpicklable:
        def __reduce__(self):
            raise TypeError("unpicklable")

    with pytest.raises(TypeError):
        empty_env.update(AddVertex(Unpicklable()))


def test_add_vertex(empty_env: Environment) -> None:
    empty_env.update(AddVertex(primitive="foo"))
    assert empty_env.find_all_of_type(str) == {"foo"}

    empty_env.commit()
    assert empty_env.find_all_of_type(str) == {"foo"}


def test_add_vertex_deduplication(empty_env: Environment) -> None:
    empty_env.update(AddVertex(primitive="foo"), AddVertex(primitive="bar"))
    vertices, _ = empty_env.find_all()
    assert {*vertices} == {"foo", "bar"}

    empty_env.update(AddVertex(primitive="foo"))
    vertices, _ = empty_env.find_all()
    assert len(vertices) == 2


def test_add_vertex_distinguishes_int_from_bool(empty_env: Environment) -> None:
    empty_env.update(AddVertex(primitive=0), AddVertex(primitive=False))
    vertices, _ = empty_env.find_all()
    assert len(vertices) == 2


def test_add_vertex_distinguishes_int_from_float(empty_env: Environment) -> None:
    empty_env.update(AddVertex(primitive=0), AddVertex(primitive=0.0))
    vertices, _ = empty_env.find_all()
    assert len(vertices) == 2


def test_add_vertex_distinguishes_str_from_str_enum(empty_env: Environment) -> None:
    empty_env.update(
        AddVertex(primitive=Biome.LAKE.value), AddVertex(primitive=Biome.LAKE)
    )
    vertices, _ = empty_env.find_all()
    assert len(vertices) == 2


def test_add_edge_without_both_vertices(empty_env: Environment) -> None:
    empty_env.update(AddEdge("from", "label", "to"))
    vertices, edges = empty_env.find_all()
    assert not vertices and not edges, "shouldn't add edge"


def test_add_edge_without_from_vertex(empty_env: Environment) -> None:
    empty_env.update(AddVertex("to"), AddEdge("from", "label", "to"))
    vertices, edges = empty_env.find_all()
    assert {*vertices} == {"to"} and not edges, "shouldn't add edge"


def test_add_edge_without_to_vertex(empty_env: Environment) -> None:
    empty_env.update(AddVertex("from"), AddEdge("from", "label", "to"))
    vertices, edges = empty_env.find_all()
    assert {*vertices} == {"from"} and not edges, "shouldn't add edge"


def test_add_edge_non_string_label(empty_env: Environment) -> None:
    empty_env.update(AddVertex("from"), AddVertex("to"))
    with pytest.raises(TypeError):
        empty_env.update(AddEdge("from", 1, "to"))


def test_add_edge(empty_env: Environment) -> None:
    empty_env.update(AddVertex("from"), AddVertex("to"), AddEdge("from", "label", "to"))
    vertices, edges = empty_env.find_all()
    assert {*vertices} == {"from", "to"} and edges == [Edge("from", "label", "to")]


def test_add_edge_deduplication(empty_env: Environment) -> None:
    empty_env.update(AddVertex("from"), AddVertex("to"), AddEdge("from", "label", "to"))
    empty_env.update(AddEdge("from", "label", "to"))
    _, edges = empty_env.find_all()
    assert edges == [Edge("from", "label", "to")]


def test_remove_nonexisting_vertex(empty_env: Environment) -> None:
    try:
        empty_env.update(RemoveVertex("foo"))
        empty_env.commit()
    except Exception as e:
        assert not e, "should not raise exception"


def test_remove_specific_vertex(empty_env: Environment) -> None:
    empty_env.update(AddVertex("foo"), AddVertex("bar"))
    empty_env.update(RemoveVertex("foo"))
    vertices, _ = empty_env.find_all()
    assert vertices == ["bar"]


def test_remove_vertex_before_commit(empty_env: Environment) -> None:
    empty_env.update(AddVertex("foo"))
    empty_env.update(RemoveVertex("foo"))

    assert_no_vertices(empty_env)

    empty_env.commit()
    assert_no_vertices(empty_env)


def test_remove_vertex_after_commit(empty_env: Environment) -> None:
    empty_env.update(AddVertex("foo"))
    empty_env.commit()
    empty_env.update(RemoveVertex("foo"))

    assert_no_vertices(empty_env)

    empty_env.commit()
    assert_no_vertices(empty_env)


def test_remove_nonexisting_edge_without_both_vertices(empty_env: Environment) -> None:
    try:
        empty_env.update(RemoveEdge("from", "label", "to"))
        empty_env.commit()
    except Exception as e:
        assert not e, "should not raise exception"


def test_remove_nonexisting_edge_without_from_vertex(empty_env: Environment) -> None:
    empty_env.update(AddVertex("to"))
    try:
        empty_env.update(RemoveEdge("from", "label", "to"))
        empty_env.commit()
    except Exception as e:
        assert not e, "should not raise exception"


def test_remove_nonexisting_edge_without_to_vertex(empty_env: Environment) -> None:
    empty_env.update(AddVertex("from"))
    try:
        empty_env.update(RemoveEdge("from", "label", "to"))
        empty_env.commit()
    except Exception as e:
        assert not e, "should not raise exception"


def test_remove_nonexisting_edge(empty_env: Environment) -> None:
    empty_env.update(AddVertex("from"), AddVertex("to"))
    try:
        empty_env.update(RemoveEdge("from", "label", "to"))
        empty_env.commit()
    except Exception as e:
        assert not e, "should not raise exception"


def test_remove_nonexisting_edge_before_commit(empty_env: Environment) -> None:
    empty_env.update(AddVertex("from"), AddVertex("to"), AddEdge("from", "label", "to"))
    empty_env.update(RemoveEdge("from", "label", "to"))
    assert_no_edges(empty_env)
    empty_env.commit()
    assert_no_edges(empty_env)


def test_remove_nonexisting_edge_after_commit(empty_env: Environment) -> None:
    empty_env.update(AddVertex("from"), AddVertex("to"), AddEdge("from", "label", "to"))
    empty_env.commit()
    empty_env.update(RemoveEdge("from", "label", "to"))
    assert_no_edges(empty_env)
    empty_env.commit()
    assert_no_edges(empty_env)


def test_remove_edges_to_all_from_nonexisting_vertex(empty_env: Environment) -> None:
    try:
        empty_env.update(RemoveEdgesToAll("from", "label"))
        empty_env.commit()
    except Exception as e:
        assert not e, "should not raise exception"


def test_remove_edges_to_all_with_nonexisting_label(empty_env: Environment) -> None:
    empty_env.update(
        AddVertex("from"), AddVertex("to"), AddEdge("from", "other_label", "to")
    )
    try:
        empty_env.update(RemoveEdgesToAll("from", "label"))
        empty_env.commit()
    except Exception as e:
        assert not e, "should not raise exception"
    _, edges = empty_env.find_all()
    assert edges == [Edge("from", "other_label", "to")]


@pytest.mark.parametrize("commit_before_removal", [False, True])
def test_remove_edges_to_all(
    empty_env: Environment, commit_before_removal: bool
) -> None:
    empty_env.update(
        AddVertex("from"),
        AddVertex("to1"),
        AddVertex("to2"),
        AddEdge("from", "label", "to1"),
        AddEdge("from", "label", "to2"),
        AddEdge("from", "other_label", "to2"),
    )
    if commit_before_removal:
        empty_env.commit()
    empty_env.update(RemoveEdgesToAll("from", "label"))
    _, edges = empty_env.find_all()
    assert edges == [Edge("from", "other_label", "to2")]
    empty_env.commit()
    _, edges = empty_env.find_all()
    assert edges == [Edge("from", "other_label", "to2")]


def assert_no_vertices(env: Environment) -> None:
    vertices, _ = env.find_all()
    assert not vertices


def assert_no_edges(env: Environment) -> None:
    _, edges = env.find_all()
    assert not edges



class SaengraError(Exception):
    """Base exception for Saengra client errors."""


class ConnectionError(SaengraError):
    """Error connecting to server."""


class CommandError(SaengraError):
    """Error executing command."""


class ConnectError(SaengraError):
    """Error connecting to graph."""


class UpdateError(CommandError):
    """Error applying updates."""


class CommitError(CommandError):
    """Error committing changes."""


class RollbackError(CommandError):
    """Error rolling back changes."""


class ObserveError(CommandError):
    """Error registering observers."""


class FindError(CommandError):
    """Error finding vertices or edges."""


class MatchError(CommandError):
    """Error matching patterns."""



import typing
from dataclasses import dataclass
from functools import reduce
from types import GenericAlias, UnionType
from typing import _UnionGenericAlias  # type: ignore
from typing import Any, ForwardRef

from _operator import or_

from saengra.utilities.typing import Module


@dataclass(frozen=True, slots=True)
class LookaheadType:
    module: Any
    type_: str

    def resolve(self) -> type:
        return getattr(self.module, self.type_)


@dataclass
class UnionAnnotation:
    types: tuple[type, ...]


@dataclass
class CollectionAnnotation:
    collection_type: type | UnionAnnotation
    item_type: type | LookaheadType


@dataclass
class OptionalAnnotation:
    type_: type | UnionAnnotation | LookaheadType


@dataclass
class ScalarAnnotation:
    type_: type | UnionAnnotation | LookaheadType


ParsedAnnotation = CollectionAnnotation | OptionalAnnotation | ScalarAnnotation


class Types:
    """This namespace makes it more convenient to write match-case statements involving types."""

    NoneType = type(None)
    frozenset = frozenset
    list = list
    set = set
    tuple = tuple


def parse_annotation(module: Module, annotation: Any) -> ParsedAnnotation:
    match annotation:
        case GenericAlias(
            __origin__=Types.list | Types.set as collection_type,
            __args__=(item_type,),
        ):
            return CollectionAnnotation(
                collection_type=collection_type,
                item_type=_parse_scalar_annotation(module, item_type),
            )

        case _UnionGenericAlias(
            __origin__=typing.Union, __args__=(item_type, Types.NoneType)  # type: ignore
        ):
            return OptionalAnnotation(type_=_parse_scalar_annotation(module, item_type))

        case UnionType(__args__=(item_type, Types.NoneType)):
            # Foo | None
            return OptionalAnnotation(type_=_parse_scalar_annotation(module, item_type))

        case UnionType(__args__=(*item_types, Types.NoneType)):
            # Foo | Bar | None
            union_without_none = reduce(or_, item_types)
            return OptionalAnnotation(
                type_=_parse_scalar_annotation(module, union_without_none)
            )

        case UnionType(__args__=item_types):
            # Foo | Bar
            return ScalarAnnotation(type_=UnionAnnotation(types=item_types))

        case ForwardRef(__forward_arg__=a):
            return ScalarAnnotation(type_=_parse_scalar_annotation(module, a))

        case item_type:
            return ScalarAnnotation(type_=_parse_scalar_annotation(module, item_type))


def _parse_scalar_annotation(
    module: Module, annotation: Any
) -> type | UnionAnnotation | LookaheadType:
    if isinstance(annotation, str):
        return LookaheadType(module=module, type_=annotation)
    if isinstance(annotation, ForwardRef):
        return LookaheadType(module=module, type_=annotation.__forward_arg__)
    if isinstance(annotation, UnionType):
        return UnionAnnotation(types=annotation.__args__)
    return annotation



from sys import intern
from typing import TYPE_CHECKING, Any, assert_never

from saengra.utilities.typing import check_alive, is_entity

if TYPE_CHECKING:
    from saengra.entity import EnvProtocol, Entity

from saengra.graph import AddEdge, AddVertex, Label, Primitive, RemoveEdgesToAll
from saengra.utilities.annotations import (
    CollectionAnnotation,
    OptionalAnnotation,
    ParsedAnnotation,
    ScalarAnnotation,
)
from saengra.utilities.containers import RelatedEntitiesSet

_object_getattribute = object.__getattribute__
_object_setattr = object.__setattr__
_object_delattr = object.__delattr__


class EntityProperty:

    def compose_init_code(self) -> tuple[list[str], dict[str, Any]]:
        """Returns lines of code to add to "Entity".__init__ method and corresponding update to locals() for exec()."""
        return [], {}

    def compose_slots(self) -> list[str]:
        """Returns the list of attribute names that are going to be used and have to be added to entity's __slots__."""
        return []

    def reset_cache(
        self, env: "EnvProtocol", e: "Entity", primitive: Primitive
    ) -> None:
        pass

    def clear_cache(self, e: "Entity") -> None:
        pass

    def add_to_cache(self, e: "Entity", what: Any):
        pass

    def remove_from_cache(self, e: "Entity", what: "Entity") -> None:
        pass


class PrimitiveProperty(EntityProperty):
    __slots__ = ("_attr_name",)

    def __init__(self, attr_name: str) -> None:
        self._attr_name = attr_name

    def __get__(self, instance: "Entity | None", owner):
        if instance is None:
            return self

        return _object_getattribute(instance.primitive, self._attr_name)

    def __delete__(self, instance: "Entity"):
        raise AttributeError(self._attr_name)


class RelatedCollectionProperty(EntityProperty):
    __slots__ = (
        "_label",
        "_cache_attr_name",
        "_item_type_annotation",
        "_proxy_type",
    )

    def __init__(self, label: Label, annotation: CollectionAnnotation) -> None:
        self._label = intern(label)
        self._cache_attr_name = _make_cache_attr_name(label)
        self._item_type_annotation = annotation.item_type
        self._proxy_type = {
            set: RelatedEntitiesSet,
        }[
            annotation.collection_type  # type: ignore
        ]

    def make_collection(self, env: "EnvProtocol", e: "Entity", primitive: Primitive):
        return self._proxy_type(env, e, primitive, self._label)

    def __get__(self, instance: "Entity | None", owner):
        if instance is None:
            return self
        return _object_getattribute(instance, self._cache_attr_name)

    def __set__(self, instance: "Entity", value: Any) -> None:
        collection = _object_getattribute(instance, self._cache_attr_name)
        collection.assign([*value])

    def __delete__(self, e: "Entity"):
        raise AttributeError(self._label)

    def compose_init_code(self) -> tuple[list[str], dict[str, Any]]:
        result = [
            f"self.{self._cache_attr_name} = cls.{self._label}.make_collection(self._env, self, self.primitive)"
        ]
        return result, {}

    def compose_slots(self) -> list[str]:
        return [self._cache_attr_name]

    def reset_cache(
        self, env: "EnvProtocol", e: "Entity", primitive: Primitive
    ) -> None:
        collection = self.make_collection(env, e, primitive)
        _object_setattr(e, self._cache_attr_name, collection)

    def clear_cache(self, e: "Entity") -> None:
        collection = _object_getattribute(e, self._cache_attr_name)
        collection._remove_inverse(collection._cached_values)
        collection._cached_values.clear()

    def add_to_cache(self, e: "Entity", what: Any):
        collection = _object_getattribute(e, self._cache_attr_name)
        collection._add_to_cache(what)

    def remove_from_cache(self, e: "Entity", what: "Entity") -> None:
        collection = _object_getattribute(e, self._cache_attr_name)
        collection._cached_values.remove(what)


class RelatedEntityPropertyBase(EntityProperty):
    __slots__ = ("_label", "_cache_attr_name")

    def __init__(
        self, label: Label, annotation: ScalarAnnotation | OptionalAnnotation
    ) -> None:
        self._label = intern(label)
        self._cache_attr_name = _make_cache_attr_name(label)

    def compose_slots(self) -> list[str]:
        return [self._cache_attr_name]

    def add_to_cache(self, e: "Entity", what: Any) -> None:
        if is_entity(prev := _object_getattribute(e, self._cache_attr_name)):
            prev._inverse.remove((e, self._label))
        _object_setattr(e, self._cache_attr_name, what)
        if is_entity(what):
            what._inverse.add((e, self._label))


class RelatedOptionalEntityProperty(RelatedEntityPropertyBase):
    __slots__ = RelatedEntityPropertyBase.__slots__

    def __get__(self, instance: "Entity | None", owner):
        if instance is None:
            return self
        return _object_getattribute(instance, self._cache_attr_name)

    def __set__(self, instance: "Entity", value: Any) -> None:
        # Update graph
        from_, label = instance.primitive, self._label
        instance._env.update(RemoveEdgesToAll(from_, label))
        if value is not None:
            if is_entity(value):
                if not value.alive:
                    raise RuntimeError(
                        f"trying to assign `.{self._label}` to {value} that is not alive"
                    )
                instance._env.update(AddEdge(from_, label, value.primitive))
            else:
                instance._env.update(AddVertex(value), AddEdge(from_, label, value))
        # Update entities
        self.add_to_cache(instance, value)

    def __delete__(self, instance: "Entity"):
        raise AttributeError(self._label)

    def compose_init_code(self) -> tuple[list[str], dict[str, Any]]:
        return [f"self.{self._cache_attr_name} = None"], {}

    def reset_cache(
        self, env: "EnvProtocol", e: "Entity", primitive: Primitive
    ) -> None:
        _object_setattr(e, self._cache_attr_name, None)

    def clear_cache(self, e: "Entity") -> None:
        _object_setattr(e, self._cache_attr_name, None)

    def remove_from_cache(self, e: "Entity", what: "Entity") -> None:
        _object_setattr(e, self._cache_attr_name, None)


class Uninitialized:
    def __str__(self) -> str:
        return "Uninitialized"

    def __repr__(self) -> str:
        return "Uninitialized"


_uninitialized = Uninitialized()


class RelatedEntityProperty(RelatedEntityPropertyBase):
    __slots__ = RelatedEntityPropertyBase.__slots__

    def __get__(self, instance: "Entity | None", owner):
        if instance is None:
            return self
        try:
            return _object_getattribute(instance, self._cache_attr_name)
        except AttributeError as exc:
            raise RuntimeError(
                f"{instance!r} doesn't have a linked {self._label!r}"
            ) from exc

    def __set__(self, instance: "Entity", value: Any) -> None:
        # Update graph
        from_, label = instance.primitive, self._label
        instance._env.update(RemoveEdgesToAll(from_, label))
        if is_entity(value):
            if not value.alive:
                raise RuntimeError(
                    f"trying to assign `.{self._label}` to {value} that is not alive"
                )
            instance._env.update(AddEdge(from_, label, value.primitive))
        else:
            instance._env.update(AddVertex(value), AddEdge(from_, label, value))
        # Update entities
        self.add_to_cache(instance, value)

    def __delete__(self, instance: "Entity"):
        raise AttributeError(self._label)

    def compose_init_code(self) -> tuple[list[str], dict[str, Any]]:
        result = [f"self.{self._cache_attr_name} = _uninitialized"]
        return result, {"_uninitialized": _uninitialized}

    def reset_cache(
        self, env: "EnvProtocol", e: "Entity", primitive: Primitive
    ) -> None:
        pass  # don't init cache, this attribute has to be set

    def clear_cache(self, e: "Entity") -> None:
        _object_setattr(e, self._cache_attr_name, _uninitialized)

    def remove_from_cache(self, e: "Entity", what: "Entity") -> None:
        _object_setattr(e, self._cache_attr_name, _uninitialized)


def related_entity_property(
    label: Label, annotation: ParsedAnnotation
) -> EntityProperty:
    match annotation:
        case CollectionAnnotation():
            return RelatedCollectionProperty(label, annotation)

        case OptionalAnnotation():
            return RelatedOptionalEntityProperty(label, annotation)

        case ScalarAnnotation():
            return RelatedEntityProperty(label, annotation)

        case other:
            assert_never(other)


def _make_cache_attr_name(label: Label) -> str:
    return intern(f"_{label}")



import sys
from dataclasses import dataclass, fields
from inspect import getmodule
from typing import Any, Protocol, Self, Type, TypeAlias, dataclass_transform

from saengra.graph import Label, Primitive, Update
from saengra.utilities.typing import check_alive


class EnvProtocol(Protocol):
    def _get_living_entity(self, primitive: Primitive) -> "Entity | None":
        pass

    def _get_temporary_entity(self, primitive: Primitive) -> "Entity | None":
        pass

    def _register_temporary_entity(
        self, primitive: Primitive, entity: "Entity"
    ) -> None:
        pass

    def add(self, *entities: "Entity") -> None:
        pass

    def remove(self, *entities: "Entity") -> None:
        pass

    def update(self, *updates: Update) -> None:
        pass

    def update_many(self, updates: list[Update]) -> None:
        pass


@dataclass_transform(eq_default=True, order_default=True, frozen_default=True)
def primitive(cls: Type) -> Type:
    result = dataclass(frozen=True, slots=True, order=True)(cls)

    # Optimized pickle functions (compared to standard library at its current state):
    result_fields = fields(result)
    if len(result_fields) == 1:
        (f,) = result_fields
        getstate_code = (
            f"def __getstate_{result.__name__}__(self): return self.{f.name}"
        )
        setstate_code = f"def __setstate_{result.__name__}__(self, state): _object_setattr(self, {f.name!r}, state)"
    elif result_fields:
        getstate_code = f"def __getstate_{result.__name__}__(self): return {', '.join(f'self.{f.name}' for f in result_fields)}"
        setstate_code = f"def __setstate_{result.__name__}__(self, state):\n  " + (
            "\n  ".join(
                f"  _object_setattr(self, {f.name!r}, state[{i}])"
                for i, f in enumerate(result_fields)
            )
        )
    else:
        # Since __setstate__ is essentially a no-op here, we can return None
        # from __getstate__ and make pickle avoid calling __setstate__ whatsoever.
        getstate_code = f"def __getstate_{result.__name__}__(self): return None"
        setstate_code = f"def __setstate_{result.__name__}__(self, state): pass"
    globals_ = {"_object_setattr": object.__setattr__}
    locals_ = {}
    exec(getstate_code, {}, locals_)
    exec(setstate_code, globals_, locals_)
    result.__getstate__ = locals_[f"__getstate_{result.__name__}__"]
    result.__setstate__ = locals_[f"__setstate_{result.__name__}__"]

    setattr(result, PRIMITIVE_MARKER_ATTR_NAME, True)
    return result


def _find_primitive_ancestor_type(name: str, mro: tuple[type, ...]) -> Type[Primitive]:
    primitives = [t for t in mro if not issubclass(t, Entity) and _is_primitive_type(t)]
    if not primitives:
        raise ValueError(f"{name}: no ancestor marked as primitive")
    if len(primitives) > 1:
        raise ValueError(
            f"{name}: too many ancestors marked as primitives: {primitives}"
        )
    return primitives[0]


def _is_primitive_type(t: type) -> bool:
    return getattr(t, PRIMITIVE_MARKER_ATTR_NAME, False)


InverseRefs: TypeAlias = set[tuple["Entity", Label]]


class EntityMeta(type):
    def __new__(mcls, name: str, bases: tuple[type, ...], ns: dict[str, Any]) -> type:
        from saengra.utilities.annotations import parse_annotation
        from saengra.utilities.properties import (
            EntityProperty,
            PrimitiveProperty,
            related_entity_property,
        )

        module = sys.modules[ns["__module__"]]
        annotations = ns.get("__annotations__", {})

        if name == "Entity" and module == getmodule(mcls):
            return super().__new__(mcls, name, bases, ns)

        primitive_type = _find_primitive_ancestor_type(name, bases)
        properties: list[EntityProperty] = []
        for attr_name in primitive_type.__annotations__:
            ns[attr_name] = property_ = PrimitiveProperty(attr_name)
            properties.append(property_)

        for attr_name, annotation in [*annotations.items()]:
            label = Label(attr_name)
            parsed_annotation = parse_annotation(module, annotation)
            property_ = related_entity_property(label, parsed_annotation)
            ns[attr_name] = property_
            properties.append(property_)

        init_cache_lines = []
        init_code_locals = {}
        slots: list[str] = [
            "__weakref__",
            "_cache_initialized",
            "_env",
            "_inverse",
            "alive",
            "primitive",
        ]
        for property_ in properties:
            extra_lines, extra_locals = property_.compose_init_code()
            init_cache_lines.extend(extra_lines)
            init_code_locals.update(extra_locals)
            slots.extend(property_.compose_slots())

        ns["__slots__"] = tuple(slots)

        if "__init__" in ns:
            raise NotImplementedError("custom __init__ is not supported yet")

        init_code_lines = [
            "self._env = env",
            "if primitive is None:",
            "    primitive, _, kwargs = self._construct_primitive(**kwargs)",
            "self.primitive = primitive",
            "if not self._cache_initialized:",
            "    cls = type(self)",
            *[f"    {line}" for line in init_cache_lines],
            "    self._cache_initialized = True",
            "for k, v in kwargs.items():",
            "    setattr(self, k, v)",
        ]
        init_code = (
            "def __init__(self, env, primitive = None, **kwargs) -> None:\n    "
            + "\n    ".join(init_code_lines)
        )
        exec(init_code, init_code_locals, ns)

        result = super().__new__(mcls, name, bases, ns)

        if result.__mro__[1] is not Entity:
            raise RuntimeError(
                "Entity must be first base class, inheriting Entity subclasses is not supported"
            )
            # We assume that Entity is the first base class to optimize `is_entity()`.

        return result


class Entity(metaclass=EntityMeta):
    """
    Sample usage:

        @primitive
        class player:
            id: UUID

        class Player(Entity, player):
            name: str
            allies: set["Player"]

        primitive = Player.primitive_type(id=uuid4())
        graph.update(AddVertex(primitive))
        player = Player(graph=graph, primitive=primitive)
    """

    __slots__ = ()

    # Class attributes:
    primitive_type: Type[Primitive]

    # Instance attributes:
    primitive: Primitive
    alive: bool
    _env: EnvProtocol
    _inverse: InverseRefs

    def __init_subclass__(cls) -> None:
        from saengra.utilities.annotations import parse_annotation
        from saengra.utilities.properties import (
            PrimitiveProperty,
            related_entity_property,
        )

        super().__init_subclass__()

        module = getmodule(cls)

        cls.primitive_type = _find_primitive_ancestor_type(cls.__name__, cls.__mro__)

        for attr_name in cls.primitive_type.__annotations__:
            setattr(cls, attr_name, PrimitiveProperty(attr_name))

        for attr_name, annotation in cls.__annotations__.items():
            label = Label(attr_name)
            parsed_annotation = parse_annotation(module, annotation)
            property_ = related_entity_property(label, parsed_annotation)
            setattr(cls, attr_name, property_)

    ####################################################################################################################

    def __new__(
        cls, env: EnvProtocol, primitive: Primitive | None = None, **kwargs
    ) -> "Entity":
        if primitive is None:
            primitive, _, kwargs = cls._construct_primitive(**kwargs)
        # noinspection PyProtectedMember
        if e := env._get_living_entity(primitive):
            return e
        # noinspection PyProtectedMember
        if e := env._get_temporary_entity(primitive):
            if kwargs:
                raise ValueError(
                    f"can't set attributes — {primitive} doesn't exist yet"
                )
            return e

        e = object.__new__(cls)
        e._inverse = set()
        e._cache_initialized = False
        e.alive = False
        # noinspection PyProtectedMember
        env._register_temporary_entity(primitive, e)
        return e

    @classmethod
    def _construct_primitive(
        cls, *args: Any, **kwargs: Any
    ) -> tuple[Primitive, tuple[Any, ...], dict[str, Any]]:
        total_expected_args = len(cls.primitive_type.__annotations__)
        matching_kwargs = {
            k: v for k, v in kwargs.items() if k in cls.primitive_type.__annotations__
        }
        matching_args = args[: total_expected_args - len(matching_kwargs)]
        primitive = cls.primitive_type(*matching_args, **matching_kwargs)
        remaining_args = args[len(matching_args) :]
        remaining_kwargs = {
            k: v
            for k, v in kwargs.items()
            if k not in cls.primitive_type.__annotations__
        }
        return primitive, remaining_args, remaining_kwargs

    ####################################################################################################################

    # Overriding mutation forbidding methods introduced by `@primitive`.
    __setattr__ = object.__setattr__
    __delattr__ = object.__delattr__

    def __hash__(self) -> int:
        return hash(self.primitive)

    def __lt__(self, other: Any) -> bool:
        match other:
            case Entity():
                return self.primitive < other.primitive  # type: ignore
        return NotImplemented

    def __eq__(self, other: Any) -> bool:
        # `other` can't be an `Entity`, because by construction of the entity
        # graph there can't be two identical entities, __new__ would return
        # an already existing one. And if we compare with `==` against the
        # same object, Python optimizes it and doesn't invoke `__eq__`.
        return self.primitive == other

    def __ne__(self, other: Any) -> bool:
        return self.primitive != other

    def __repr__(self) -> str:
        result = repr(self.primitive)
        prefix = self.primitive_type.__name__
        if result.startswith(prefix):
            return type(self).__name__ + result.removeprefix(prefix)
        return f"{type(self).__name__}({result})"

    ####################################################################################################################

    @classmethod
    def get(
        cls: Type[Self],
        env: EnvProtocol,
        primitive: Primitive | None = None,
        *,
        __allow_extra_kwargs: bool = False,
        **kwargs: object,
    ) -> Self:
        if primitive is None:
            primitive, _, kwargs = cls._construct_primitive(**kwargs)
        if kwargs and not __allow_extra_kwargs:
            raise ValueError("not supposed to add edges in .get()")
        # noinspection PyProtectedMember
        if e := env._get_living_entity(primitive):
            return e
        raise ValueError(f"{primitive} doesn't exist yet")

    @classmethod
    def get_or_none(
        cls: Type[Self],
        env: EnvProtocol,
        primitive: Primitive | None = None,
        *,
        __allow_extra_kwargs: bool = False,
        **kwargs,
    ) -> Self | None:
        if primitive is None:
            primitive, _, kwargs = cls._construct_primitive(**kwargs)
        if kwargs and not __allow_extra_kwargs:
            raise ValueError("not supposed to add edges in .get_or_none()")
        # noinspection PyProtectedMember
        if e := env._get_living_entity(primitive):
            return e
        return None

    @classmethod
    def get_or_create(
        cls: Type[Self],
        env: EnvProtocol,
        primitive: Primitive | None = None,
        *,
        defaults: dict[str, Any] | None = None,
        **kwargs,
    ) -> Self:
        if primitive is None:
            primitive, _, kwargs = cls._construct_primitive(**kwargs)
        result = cls(env=env, primitive=primitive)
        just_created = False
        if not result.alive:
            just_created = True
            env.add(result)

        for k, v in kwargs.items():
            if just_created and v is None:
                continue
            setattr(result, k, check_alive(k, v))

        if just_created and defaults:
            for k, v in defaults.items():
                if v is not None:
                    setattr(result, k, check_alive(k, v))

        return result

    @classmethod
    def create(
        cls: Type[Self], env: EnvProtocol, primitive: Primitive | None = None, **kwargs
    ) -> Self:
        if primitive is None:
            primitive, _, kwargs = cls._construct_primitive(**kwargs)
        result = cls(env=env, primitive=primitive)
        if result.alive:
            raise ValueError(f"{primitive} already exists")
        env.add(result)

        for k, v in kwargs.items():
            if v is None:
                continue
            setattr(result, k, check_alive(k, v))

        return result

    def update(self, **kwargs) -> Self:
        for k, v in kwargs.items():
            setattr(self, k, check_alive(k, v))
        return self

    def discard(self) -> None:
        if not self.alive:
            return
        self._env.remove(self)

    def remove(self) -> None:
        if not self.alive:
            raise ValueError(f"{self.primitive} doesn't exist")
        self._env.remove(self)

    def _after_add(self) -> None:
        # Invoked by Environment if after the vertex is added.
        self.alive = True

    def _after_remove(self) -> None:
        # Invoked by Environment if after the vertex is removed.
        self.alive = False
        for label in self.__annotations__:
            prop = getattr(type(self), label)
            if isinstance(related_entity := getattr(self, label, None), Entity):
                related_entity._inverse.remove((self, label))
            prop.clear_cache(self)
        for related_entity, label in self._inverse:
            prop = getattr(type(related_entity), label)
            prop.remove_from_cache(related_entity, self)
        self._inverse.clear()

    def _sync(self, edges: list[tuple[Label, "Primitive | Entity"]]) -> None:
        self._reset_cache(self._env, self.primitive)
        for label, to in edges:
            attr_name = str(label)
            prop = getattr(type(self), attr_name)
            prop.add_to_cache(self, to)

    def _reset_cache(self, env: EnvProtocol, primitive: Primitive) -> None:
        cls = type(self)
        for attr_name in cls.__annotations__:
            prop = getattr(cls, attr_name)
            prop.reset_cache(env, self, primitive)


PRIMITIVE_MARKER_ATTR_NAME = "__saengra_primitive__"

import saengra.utilities.typing

saengra.utilities.typing.Entity = Entity



from collections import defaultdict
from typing import Callable, Iterable, TypeVar

T = TypeVar("T")
K = TypeVar("K")


def group_by(what: Iterable[T], key: Callable[[T], K]) -> defaultdict[K, list[T]]:
    result: defaultdict[K, list[T]] = defaultdict(list)
    for item in what:
        item_key = key(item)
        result[item_key].append(item)
    return result



from dataclasses import dataclass, field
from enum import Enum

from saengra import messages_pb2
from saengra.graph import Edge, Primitive, Subgraph


@dataclass(frozen=True)
class Observer:
    id: str
    expression: str
    placeholder_values: tuple[Primitive, ...] = field(default_factory=tuple)
    on_create: bool = False
    on_change: bool = False
    on_delete: bool = False


class ObservationType(Enum):
    ON_CREATE = messages_pb2.CommitResponse.Observation.ON_CREATE
    ON_CHANGE = messages_pb2.CommitResponse.Observation.ON_CHANGE
    ON_DELETE = messages_pb2.CommitResponse.Observation.ON_DELETE


@dataclass(frozen=True)
class Observation:
    observer_id: str
    type: ObservationType
    variables: dict[str, Primitive]


class CommitStatus(Enum):
    OK = messages_pb2.CommitResponse.OK
    REQUIRES_REPEATED_COMMIT = messages_pb2.CommitResponse.REQUIRES_REPEATED_COMMIT


@dataclass(frozen=True)
class CommitResponse:
    status: CommitStatus
    observations: list[Observation]


@dataclass(frozen=True)
class FindResponse:
    vertices: list[Primitive]
    edges: list[Edge]


@dataclass(frozen=True)
class MatchResponse:
    subgraphs: list[Subgraph]



from termcolor import colored


def blue(s: str) -> str:
    return colored(s, "blue", force_color=True)


def cyan(s: str) -> str:
    return colored(s, "cyan", force_color=True)


def green(s: str) -> str:
    return colored(s, "green", force_color=True)


def light_green(s: str) -> str:
    return colored(s, "light_green", force_color=True)


def magenta(s: str) -> str:
    return colored(s, "magenta", force_color=True)


def red(s: str) -> str:
    return colored(s, "red", force_color=True)


def bold_red(s: str) -> str:
    return colored(s, "red", attrs=["bold"], force_color=True)


def white(s: str) -> str:
    return colored(s, "white", force_color=True)


def yellow(s: str) -> str:
    return colored(s, "yellow", force_color=True)



from saengra import primitive, Entity, Environment


@primitive
class foo:
    id: int


class Foo(Entity, foo):
    pass


@primitive
class bar:
    id: int


class Bar(Entity, bar):
    pass


@primitive
class container:
    id: int


class Container(Entity, container):
    first: Foo | Bar | None
    contains: set[Foo | Bar]


def test_optional_union():
    env = Environment(entity_types=[Foo, Bar, Container])

    c = Container.create(env, id=1)
    assert c.first is None
    assert c.contains == set()

    c.first = Foo.create(env, id=1)
    c.contains.add(c.first)
    c.contains.add(Bar.create(env, id=1))

    assert c.first == Foo.get(env, id=1)
    assert c.contains == {Foo.get(env, id=1), Bar.get(env, id=1)}



import pickle

from saengra import messages_pb2
from saengra.api import (
    CommitResponse,
    CommitStatus,
    Edge,
    FindResponse,
    MatchResponse,
    Observation,
    ObservationType,
    Observer,
)
from saengra.graph import (
    AddEdge,
    AddVertex,
    Position,
    Primitive,
    Refs,
    RemoveEdge,
    RemoveEdgesToAll,
    RemoveVertex,
    Subgraph,
    Update,
)

ApplyUpdates_Update_UpdateKind = messages_pb2.ApplyUpdates.Update.UpdateKind


class Converter:
    def __init__(self):
        self._pickle_cache: dict[Primitive, bytes] = {}

    def vertex_to_proto(self, vertex: Primitive, proto: messages_pb2.Vertex) -> None:
        # The eternal cache is present because it's very hard to normalize
        # Python objects in such a way that x == y is equivalent to
        # pickle.dumps(x) == pickle.dumps(y). Notable examples:
        # * integral floats should be normalized to ints
        # * frozendicts should be sorted by keys
        # * all of the above applies to nested elements in dataclasses, frozensets, frozendicts, etc.
        key = type(vertex), vertex
        if (value := self._pickle_cache.get(key)) is None:
            value = self._pickle_cache[key] = pickle.dumps(vertex)

        proto.type_name = type(vertex).__name__.encode("ascii")
        proto.value = value

    def vertex_from_proto(self, proto: messages_pb2.Vertex) -> Primitive:
        return pickle.loads(proto.value)

    def edge_from_proto(self, proto: messages_pb2.Edge) -> Edge:
        return Edge(
            from_=self.vertex_from_proto(getattr(proto, "from")),
            label=proto.label.decode("utf-8"),
            to=self.vertex_from_proto(proto.to),
        )

    def update_to_proto(
        self, update: Update, proto: messages_pb2.ApplyUpdates.Update
    ) -> None:
        match update:
            case AddVertex(primitive=v):
                proto.kind = ApplyUpdates_Update_UpdateKind.ADD_VERTEX
                self.vertex_to_proto(v, getattr(proto, "from"))
            case AddEdge(from_=from_, label=label, to=to):
                proto.kind = ApplyUpdates_Update_UpdateKind.ADD_EDGE
                self.vertex_to_proto(from_, getattr(proto, "from"))
                proto.label = label
                self.vertex_to_proto(to, proto.to)
            case RemoveVertex(primitive=v):
                proto.kind = ApplyUpdates_Update_UpdateKind.REMOVE_VERTEX
                self.vertex_to_proto(v, getattr(proto, "from"))
            case RemoveEdge(from_=from_, label=label, to=to):
                proto.kind = ApplyUpdates_Update_UpdateKind.REMOVE_EDGE
                self.vertex_to_proto(from_, getattr(proto, "from"))
                proto.label = label
                self.vertex_to_proto(to, proto.to)
            case RemoveEdgesToAll(from_=v, label=label):
                proto.kind = ApplyUpdates_Update_UpdateKind.REMOVE_EDGES_TO_ALL
                self.vertex_to_proto(v, getattr(proto, "from"))
                proto.label = label
            case _:
                raise TypeError(f"not an update: {update}")

    def observer_to_proto(
        self, observer: Observer, proto: messages_pb2.Observe.Observer
    ) -> None:
        proto.id = observer.id
        proto.expression = observer.expression
        for v in observer.placeholder_values:
            self.vertex_to_proto(v, proto.placeholder_values.add())
        proto.on_create = observer.on_create
        proto.on_change = observer.on_change
        proto.on_delete = observer.on_delete

    def observer_from_proto(self, proto: messages_pb2.Observe.Observer) -> Observer:
        return Observer(
            id=proto.id,
            expression=proto.expression,
            placeholder_values=tuple(
                self.vertex_from_proto(v) for v in proto.placeholder_values
            ),
            on_create=proto.on_create,
            on_change=proto.on_change,
            on_delete=proto.on_delete,
        )

    def observation_from_proto(
        self,
        proto: messages_pb2.CommitResponse.Observation,
    ) -> Observation:
        return Observation(
            observer_id=proto.observer_id,
            type=ObservationType(proto.type),
            variables={
                k: self.vertex_from_proto(v) for k, v in proto.variables.items()
            },
        )

    def commit_response_from_proto(
        self, proto: messages_pb2.CommitResponse
    ) -> CommitResponse:
        return CommitResponse(
            status=CommitStatus(proto.status),
            observations=[
                self.observation_from_proto(obs) for obs in proto.observation
            ],
        )

    def find_response_from_proto(
        self, proto: messages_pb2.FindResponse
    ) -> FindResponse:
        return FindResponse(
            vertices=[self.vertex_from_proto(v) for v in proto.vertices],
            edges=[self.edge_from_proto(e) for e in proto.edges],
        )

    def position_from_proto(self, proto: messages_pb2.Position) -> Position:
        point = "." if proto.kind == messages_pb2.Position.CORE else "o"
        return Position(
            vertex=self.vertex_from_proto(proto.vertex),
            point=point,
        )

    def subgraph_from_proto(self, proto: messages_pb2.Subgraph) -> Subgraph:
        return Subgraph(
            start_position=self.position_from_proto(proto.start_position),
            end_positions=frozenset(
                self.position_from_proto(p) for p in proto.end_positions
            ),
            vertices=frozenset(self.vertex_from_proto(v) for v in proto.vertices),
            edges=frozenset(self.edge_from_proto(e) for e in proto.edges),
            refs=Refs({k: self.vertex_from_proto(v) for k, v in proto.refs.items()}),
        )

    def match_response_from_proto(
        self, proto: messages_pb2.MatchResponse
    ) -> MatchResponse:
        return MatchResponse(
            subgraphs=[self.subgraph_from_proto(sg) for sg in proto.subgraphs],
        )



from saengra import Environment
from saengra.graph import AddVertex, RemoveVertex, AddEdge, Edge, RemoveEdge


def test_rollback_adding_vertex(empty_env: Environment) -> None:
    empty_env.update(AddVertex(1), AddVertex(2))
    empty_env.commit()

    empty_env.update(AddVertex(3))
    empty_env.rollback()
    vertices, _ = empty_env.find_all()
    assert {*vertices} == {1, 2}


def test_rollback_removing_vertex(empty_env: Environment) -> None:
    empty_env.update(AddVertex(1), AddVertex(2), AddVertex(3))
    empty_env.commit()

    empty_env.update(RemoveVertex(2))
    empty_env.rollback()
    vertices, _ = empty_env.find_all()
    assert {*vertices} == {1, 2, 3}


def test_rollback_adding_edge(empty_env: Environment) -> None:
    empty_env.update(AddVertex(1), AddVertex(2), AddVertex(3), AddEdge(1, "label", 2))
    empty_env.commit()

    empty_env.update(AddEdge(1, "label", 3))
    empty_env.rollback()
    _, edges = empty_env.find_all()
    assert edges == [Edge(1, "label", 2)]


def test_rollback_removing_edge(empty_env: Environment) -> None:
    empty_env.update(AddVertex(1), AddVertex(2), AddVertex(3), AddEdge(1, "label", 2))
    empty_env.commit()

    empty_env.update(RemoveEdge(1, "label", 2))
    empty_env.rollback()
    _, edges = empty_env.find_all()
    assert edges == [Edge(1, "label", 2)]



from typing import Iterable, Protocol, TypeAlias

from saengra.api import Observation, Observer
from saengra.graph import Edge, Primitive, Subgraph, Update

ShouldCommitAgain: TypeAlias = bool


class Adapter(Protocol):
    """Protocol for adapters that communicate with the saengra graph engine."""

    def update(
        self, update_or_updates: Update | list[Update] | tuple[Update, ...]
    ) -> None: ...

    def flush(self) -> None: ...

    def commit(self) -> tuple[ShouldCommitAgain, list[Observation]]: ...

    def rollback(self) -> None: ...

    def find_vertices(
        self, *, with_type_name: str | None = None
    ) -> list[Primitive]: ...

    def find_edges(
        self,
        *,
        with_label: str | None = None,
        from_vertices: list[Primitive] | tuple[Primitive, ...] | None = None,
        to_vertices: list[Primitive] | tuple[Primitive, ...] | None = None,
    ) -> list[Edge]: ...

    def find_all(self) -> tuple[list[Primitive], list[Edge]]: ...

    def observe(self, observers: Iterable[Observer]) -> None: ...

    def match(
        self, expression: str, *placeholder_values: Primitive
    ) -> list[Subgraph]: ...



from typing import Any, Generic, Type, TypeVar

from saengra.entity import Entity
from saengra.environment import Environment

T = TypeVar("T", bound=Entity)


class EntityFactory(Generic[T]):
    def __init__(
        self,
        entity_type: Type[T],
        env: Environment,
        *,
        defaults: dict[str, Any] | None = None,
        **kwargs,
    ) -> None:
        self._entity_type = entity_type
        self._env = env
        self._defaults = defaults
        self._kwargs = kwargs

    def get(self) -> T:
        return self._entity_type.get(
            env=self._env, _Entity__allow_extra_kwargs=True, **self._kwargs
        )

    def get_or_none(self) -> T | None:
        return self._entity_type.get_or_none(
            env=self._env, _Entity__allow_extra_kwargs=True, **self._kwargs
        )

    def get_or_create(self) -> T:
        return self._entity_type.get_or_create(
            env=self._env, defaults=self._defaults, **self._kwargs
        )

    def create(self) -> T:
        return self._entity_type.create(env=self._env, **self._kwargs)

    def discard(self) -> None:
        if e := self.get_or_none():
            e.discard()



import os
import sys
import subprocess
import shutil
from pathlib import Path
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def build_extension(self, ext):
        if not isinstance(ext, CMakeExtension):
            super().build_extension(ext)
            return

        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))

        if not extdir.endswith(os.path.sep):
            extdir += os.path.sep

        if "CMAKE_BUILD_PARALLEL_LEVEL" not in os.environ:
            import multiprocessing

            os.environ["CMAKE_BUILD_PARALLEL_LEVEL"] = str(multiprocessing.cpu_count())

        import sysconfig as _sysconfig

        python_include = _sysconfig.get_path("include")

        cmake_args = [
            f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={extdir}",
            f"-DPython_EXECUTABLE={sys.executable}",
            f"-DPython_ROOT_DIR={sys.prefix}",
            f"-DPython_INCLUDE_DIR={python_include}",
            f"-DPython_FIND_STRATEGY=LOCATION",
            f"-DCMAKE_BUILD_TYPE={'RelWithDebInfo' if os.environ.get('SAENGRA_DEBUG') else 'Release'}",
            f"-DSAENGRA_STATIC_DEPS={'ON' if os.environ.get('SAENGRA_STATIC_DEPS') else 'OFF'}",
        ]

        build_args = []

        if sys.platform.startswith("darwin"):
            homebrew_prefix = self.get_homebrew_prefix()
            if homebrew_prefix:
                cmake_args.extend(
                    [
                        f"-DCMAKE_PREFIX_PATH={homebrew_prefix}",
                    ]
                )

            if "MACOSX_DEPLOYMENT_TARGET" not in os.environ:
                cmake_args.append("-DCMAKE_OSX_DEPLOYMENT_TARGET=10.15")

        if not self.debug:
            cmake_args.append("-DCMAKE_BUILD_TYPE=Release")

        build_temp = Path(self.build_temp) / ext.name
        if not build_temp.exists():
            build_temp.mkdir(parents=True)

        subprocess.check_call(["cmake", ext.sourcedir] + cmake_args, cwd=build_temp)

        subprocess.check_call(["cmake", "--build", "."] + build_args, cwd=build_temp)

        import sysconfig

        ext_suffix = sysconfig.get_config_var("EXT_SUFFIX")
        if ext_suffix:
            lib_dir = Path(extdir)
            if lib_dir.exists():
                source_lib = lib_dir / f"{ext.name.split('.')[-1]}.so"
                if not source_lib.exists():
                    source_lib = lib_dir / ext.name.split(".")[-1]

                if source_lib.exists():
                    target_lib = lib_dir / f"{ext.name.split('.')[-1]}{ext_suffix}"
                    if source_lib != target_lib:
                        shutil.copy2(source_lib, target_lib)

    @staticmethod
    def get_homebrew_prefix():
        try:
            result = subprocess.run(
                ["brew", "--prefix"], capture_output=True, text=True, check=True
            )
            return result.stdout.strip()
        except (subprocess.CalledProcessError, FileNotFoundError):
            return None


setup(
    ext_modules=[CMakeExtension("saengra.c_extension")],
    cmdclass={"build_ext": CMakeBuild},
    zip_safe=False,
)



"""Saengra Python client library."""

__version__ = "0.1.0"

from saengra.c_extension import DirectAdapter
from saengra.entity import Entity, primitive
from saengra.environment import Environment
from saengra.factory import EntityFactory
from saengra.observer import observer
from saengra.socket_adapter import SocketAdapter



from typing import Any, Protocol, TypeAlias

from saengra.api import ObservationType
from saengra.api import Observer as APIObserver
from saengra.graph import Primitive


def observer(
    expr: str, *placeholder_values: Primitive, dont_register: bool = False
) -> "Observer":
    return Observer(expr, placeholder_values, dont_register=dont_register)


def collect_all(*modules: Any) -> list["Observer"]:
    collected_observers: set[Observer] = set()
    for module in modules:
        for value in module.__dict__.values():
            observers = getattr(value, _OBSERVERS_ATTR_NAME, [])
            collected_observers.update(observers)
    return _deduplicate_observers(collected_observers)


def _deduplicate_observers(observers: set["Observer"]) -> list["Observer"]:
    # It is very unlikely, but possible, that the same expression (with the same placeholders)
    # is watched more than once. That would lead to computational overhead, as there will be
    # more than one corresponding incremental matcher. The situation might also be surprising
    # for debug, so, since this is only invoked during the creation of a graph, we do the work.
    deduplicated_observers: dict[tuple[str, tuple[Primitive, ...]], Observer] = {}

    used_ids: set[str] = set()

    def compose_id(expr: str) -> str:
        # TODO maybe replace question marks with placeholder value reprs
        id_ = base = _generate_id_from_expression(expr)
        i = 1
        while id_ in used_ids:
            i += 1
            id_ = f"#{i}: {base}"
        used_ids.add(id_)
        return id_

    for w in observers:
        key = (w.expression, w.placeholder_values)
        if key not in deduplicated_observers:
            id_ = compose_id(w.expression)
            deduplicated_observers[key] = Observer(
                w.expression, w.placeholder_values
            ).with_id(id_)
        dw = deduplicated_observers[key]
        dw.create_handlers.extend(w.create_handlers)
        dw.change_handlers.extend(w.change_handlers)
        dw.delete_handlers.extend(w.delete_handlers)

    return [
        dw
        for dw in deduplicated_observers.values()
        if dw.create_handlers or dw.change_handlers or dw.delete_handlers
    ]


class RefsHandler(Protocol):
    @property
    def __name__(self) -> str:
        pass

    def __call__(self, *args, **kwargs) -> None:
        pass


CreateHandler: TypeAlias = RefsHandler
DeleteHandler: TypeAlias = RefsHandler
ChangeHandler: TypeAlias = RefsHandler


class Observer:
    def __init__(
        self,
        expr: str,
        placeholder_values: tuple[Primitive, ...],
        *,
        dont_register: bool = False,
    ) -> None:
        self.id = _generate_id_from_expression(expr)
        self.expression = expr
        self.placeholder_values = placeholder_values
        self.create_handlers: list[CreateHandler] = []
        self.change_handlers: list[ChangeHandler] = []
        self.delete_handlers: list[DeleteHandler] = []
        self.dont_register = dont_register

    def with_id(self, id_: str) -> "Observer":
        self.id = id_
        return self

    def on_create(self, handler: CreateHandler) -> CreateHandler:
        self.create_handlers.append(handler)
        if self.dont_register:
            return handler
        if getattr(handler, _OBSERVERS_ATTR_NAME, None) is None:
            handler.__observers__ = []  # type: ignore
        if self not in handler.__observers__:  # type: ignore
            handler.__observers__.append(self)  # type: ignore
        return handler

    def on_change(self, handler: ChangeHandler) -> ChangeHandler:
        self.change_handlers.append(handler)
        if self.dont_register:
            return handler
        if getattr(handler, _OBSERVERS_ATTR_NAME, None) is None:
            handler.__observers__ = []  # type: ignore
        if self not in handler.__observers__:  # type: ignore
            handler.__observers__.append(self)  # type: ignore
        return handler

    def on_create_or_change(self, handler: RefsHandler) -> RefsHandler:
        handler = self.on_create(handler)
        handler = self.on_change(handler)
        return handler

    def on_delete(self, handler: DeleteHandler) -> DeleteHandler:
        self.delete_handlers.append(handler)
        if self.dont_register:
            return handler
        if getattr(handler, _OBSERVERS_ATTR_NAME, None) is None:
            handler.__observers__ = []  # type: ignore
        if self not in handler.__observers__:  # type: ignore
            handler.__observers__.append(self)  # type: ignore
        return handler

    def to_api(self) -> APIObserver:
        return APIObserver(
            id=self.id,
            expression=self.expression,
            placeholder_values=self.placeholder_values,
            on_create=bool(self.on_create),
            on_change=bool(self.on_change),
            on_delete=bool(self.on_delete),
        )

    def pick_handlers(self, observation_type: ObservationType) -> list[RefsHandler]:
        match observation_type:
            case ObservationType.ON_CREATE:
                return self.create_handlers
            case ObservationType.ON_CHANGE:
                return self.change_handlers
            case ObservationType.ON_DELETE:
                return self.delete_handlers
            case _:
                raise ValueError(f"unknown observation type: {observation_type!r}")


def _generate_id_from_expression(expr: str) -> str:
    expr = expr.replace("--", "-")
    expr = expr.replace("<-", "<")
    expr = expr.replace("->", ">")
    expr = expr.replace(" as ", "$as$")
    expr = " ".join(expr.split())
    expr = expr.replace(" )", ")")
    expr = expr.replace("$as$", " as ")
    return expr


_OBSERVERS_ATTR_NAME = "__observers__"



import sys
from typing import Any, TypeAlias

Module: TypeAlias = type(sys)  # type: ignore[valid-type]

Entity: type = None  # type: ignore[assignment]
# This is filled later in entity.py.
# Used to avoid circular imports.
# And this is way faster than doing local import in every `is_entity()` call (~2x speedup).


def is_entity(obj: object) -> bool:
    # return isinstance(obj, Entity)
    return type(obj).__mro__[1] is Entity
    # This optimization actually saves 50% of time spent on this function calls
    # (and there are normally LOTS of calls, up to 1% of total runtime).


def check_alive(k: str, v: Any) -> Any:
    if is_entity(v) and not v.alive:
        raise RuntimeError(f"trying to assign `.{k}` to {v} that is not alive")
    return v
